@@ -21,9 +21,18 @@
 #include <nlist.h>
 
 #ifdef IRIX
+#  ifdef IRIX62
+#    if _MIPS_SZPTR == 64
+#       define _K64U64 1
+#    else
+#       define _K32U32 1
+#    endif
+#  endif
 #  include <sys/sbd.h>
-#  ifdef IRIX6
+#  ifndef K0_TO_PHYS
 #    define K0_TO_PHYS(x)   (x) 
+#  endif
+#  if defined(IRIX6) || (defined(IRIX62) && _MIPS_SZPTR == 64)
 #    define nlist nlist64
 #  endif
 #endif
@@ -162,11 +171,8 @@ int kvm_nlist(kd, nl)
   int code;
   int i;
 
-#ifdef IRIX6
-  code = nlist64(kd->namelist, nl);
-#else
   code = nlist(kd->namelist, nl);
-#endif
+
   if (code != 0)
     return code;
   
@@ -211,10 +217,19 @@ static int readbuf(fd, addr, buf, len)
   addr = K0_TO_PHYS(addr);
 #endif
   errno = 0;
-  if (lseek(fd, addr, 0) == -1 && errno != 0)
-    return -1;
-
+#if defined(__alpha)
+  /*
+   * Let us be paranoid about return values.
+   * It should be like this on all implementations,
+   * but some may have broken lseek or read returns.
+   */
+  if (lseek(fd, addr, 0) != addr || errno != 0) return -1;
+  if (read(fd, buf, len) != len  || errno != 0) return -1;
+  return len;
+#else
+  if (lseek(fd, addr, 0) == -1 && errno != 0) return -1;
   return read(fd, buf, len);
+#endif
 }
 
 int kvm_read(kd, addr, buf, len)
