@@ -2,7 +2,7 @@
 ** parse.c                         This file contains the protocol parser
 **
 ** This program is in the public domain and may be used freely by anyone
-** who wants to. 
+** who wants to.
 **
 ** Last update: 23 Feb 1994
 **
@@ -17,6 +17,9 @@
 #include <errno.h>
 #include <ctype.h>
 #include <pwd.h>
+#ifdef ALLOW_FORMAT
+#  include <grp.h>
+#endif
 
 #include <sys/types.h>
 #include <netinet/in.h>
@@ -38,7 +41,7 @@
 extern int errno;
 #endif
 
-#if defined(SOLARIS) || defined(__linux__)
+#ifdef SOLARIS
 #  include <string.h>
 #  include <stdlib.h>
 #endif
@@ -59,7 +62,7 @@ static int eat_whitespace()
 {
   int c;
 
-  
+
   while ((c = getchar()) != EOF &&
 	 !(c == '\r' || c == '\n'))
     ;
@@ -71,7 +74,7 @@ static int eat_whitespace()
 
   if (c != EOF)
     ungetc(c, stdin);
-  
+
   return (c != EOF);
 }
 
@@ -108,10 +111,10 @@ static int valid_fhost(faddr, password)
       }
     }
   }
-      
+
   if (indirect_password == NULL)
     return 1;
-  
+
   if (strcmp(password, indirect_password))
   {
     syslog(LOG_NOTICE, "valid_fhost: Invalid password from: %s",
@@ -133,11 +136,11 @@ static int check_noident(homedir)
   char *tmp_path;
   struct stat sbuf;
   int rcode;
-  
+
 
   if (!homedir)
     return 0;
-  
+
   tmp_path = (char *) malloc(strlen(homedir) + sizeof("/.noident") + 1);
   if (!tmp_path)
     return 0;
@@ -158,6 +161,12 @@ int parse(fp, laddr, faddr)
 {
   int uid, try, rcode;
   struct passwd *pwp;
+#ifdef ALLOW_FORMAT
+  int pid;
+  char *cmd, *cmd_and_args;
+  struct group *grp;
+  char grname[128];
+#endif
   char lhostaddr[16];
   char fhostaddr[16];
   char password[33];
@@ -170,11 +179,15 @@ int parse(fp, laddr, faddr)
 #endif
   struct in_addr laddr2;
   struct in_addr faddr2;
-  
-  
+  int k_opened;
+
+  k_opened = 0;
+
+
   if (debug_flag && syslog_flag)
     syslog(LOG_DEBUG, "In function parse()");
-  
+
+
   /*
   ** Get the local/foreign port pair from the luser
   */
@@ -182,13 +195,13 @@ int parse(fp, laddr, faddr)
   {
     if (debug_flag && syslog_flag)
       syslog(LOG_DEBUG, "  Before fscanf()");
-    
+
     faddr2 = *faddr;
     laddr2 = *laddr;
     lport = fport = 0;
     lhostaddr[0] = fhostaddr[0] = password[0] = '\0';
 
-    
+
     /* Read query from client */
     rcode = fscanf(fp, " %d , %d", &lport, &fport);
 
@@ -203,11 +216,11 @@ int parse(fp, laddr, faddr)
       /* Skip leading space up to EOF, EOL or non-space char */
       while ((c = getc(fp)) == ' ' || c == '\t')
 	;
-      
+
       if (rcode <= 0)
       {
 #ifdef STRONG_LOG
-	  if (syslog_flag)
+	if (syslog_flag)
 	      syslog(LOG_NOTICE, "from: %s (%s) INVALID REQUEST",
 		     inet_ntoa(*faddr), gethost(faddr));
 #endif
@@ -254,7 +267,7 @@ int parse(fp, laddr, faddr)
 		syslog(LOG_NOTICE, "from: %s (%s) INVALID PROXY REQUEST",
 		       inet_ntoa(*faddr), gethost(faddr));
 #endif
-	    
+
 	  printf("%d , %d : ERROR : %s\r\n",
 		 lport, fport,
 		 unknown_flag ? "UNKNOWN-ERROR" : "X-INVALID-REQUEST");
@@ -263,7 +276,7 @@ int parse(fp, laddr, faddr)
 
 	if (rcode == 4)
 	  laddr2.s_addr = inet_addr(lhostaddr);
-	
+
 	faddr2.s_addr = inet_addr(fhostaddr);
 
 #ifdef STRONG_LOG
@@ -274,17 +287,17 @@ int parse(fp, laddr, faddr)
 	    strcpy(a1, inet_ntoa(*faddr));
 	    strcpy(a2, inet_ntoa(faddr2));
 	    strcpy(a3, inet_ntoa(laddr2));
-		   
+
 	    syslog(LOG_NOTICE,
 		   "from: %s (%s) PROXY REQUEST for %d, %d between %s and %s",
 		   a1, gethost(faddr), lport, fport, a2, a3);
 	}
 #endif
-		   
+
 	proxy(&laddr2, &faddr2, lport, fport, NULL);
 	continue;
       }
-      
+
       /*
       ** Non-standard extended remote indirect request
       */
@@ -292,7 +305,7 @@ int parse(fp, laddr, faddr)
       {
 	/* We have a colon char, check for port numbers */
 	rcode = fscanf(fp, " %d , %d", &lport, &fport);
-	
+
 	/* Skip leading space up to EOF, EOL or non-space char */
 	while ((c = getc(fp)) == ' ' || c == '\t')
 	  ;
@@ -304,13 +317,13 @@ int parse(fp, laddr, faddr)
 		syslog(LOG_NOTICE, "from: %s (%s) INVALID REMOTE REQUEST",
 		       inet_ntoa(*faddr), gethost(faddr));
 #endif
-	    
+
 	  printf("%d , %d : ERROR : %s\r\n",
 		 lport, fport,
 		 unknown_flag ? "UNKNOWN-ERROR" : "X-INVALID-REQUEST");
 	  continue;
 	}
-	    
+
 	/* We have a colon char, check for addr and password */
 	rcode = fscanf(fp, " %15[0-9.] , %32[^ \t\r\n]",
 		       fhostaddr, password);
@@ -329,7 +342,7 @@ int parse(fp, laddr, faddr)
 		   unknown_flag ? "UNKNOWN-ERROR" : "X-INVALID-REQUEST");
 	    continue;
 	}
-	
+
 	/*
 	** Verify that the host originating the indirect request
 	** is allowed to do that
@@ -348,7 +361,7 @@ int parse(fp, laddr, faddr)
 		   unknown_flag ? "UNKNOWN-ERROR" : "X-ACCESS-DENIED");
 	    continue;
 	}
-	
+
 	faddr2.s_addr = inet_addr(fhostaddr);
 #ifdef STRONG_LOG
 	if (syslog_flag)
@@ -356,15 +369,15 @@ int parse(fp, laddr, faddr)
 	    char a1[64];
 
 	    strcpy(a1, inet_ntoa(*faddr));
-	    
-	    syslog(LOG_NOTICE,
+
+	    syslog(LOG_INFO,
 	   "from: %s (%s) REMOTE REQUEST for %d, %d from %s with password %s",
 		   a1, gethost(faddr), lport, fport,
 		   inet_ntoa(faddr2), password);
 	}
 #endif
     }
-      
+
       else
       {
 #ifdef STRONG_LOG
@@ -372,7 +385,7 @@ int parse(fp, laddr, faddr)
 	      syslog(LOG_NOTICE, "from: %s (%s) UNKNOWN REQUEST: %s",
 		     inet_ntoa(*faddr), gethost(faddr), arg);
 #endif
-	  
+
 	  printf("%d , %d : ERROR : %s\r\n",
 		 lport, fport,
 		 unknown_flag ? "UNKNOWN-ERROR" : "X-INVALID-REQUEST");
@@ -380,7 +393,7 @@ int parse(fp, laddr, faddr)
       }
     }
 #endif /* EXTENSIONS */
-    
+
     if (rcode < 2 || lport < 1 || lport > 65535 || fport < 1 || fport > 65535)
     {
 #ifdef STRONG_LOG
@@ -407,7 +420,7 @@ int parse(fp, laddr, faddr)
 	    syslog(LOG_NOTICE, "scanf: invalid-port(s): %d , %d from %s",
 		   lport, fport, gethost(faddr));
 #endif
-      
+
       printf("%d , %d : ERROR : %s\r\n",
 	     lport, fport,
 	     unknown_flag ? "UNKNOWN-ERROR" : "INVALID-PORT");
@@ -417,33 +430,94 @@ int parse(fp, laddr, faddr)
 #ifdef STRONG_LOG
       if (syslog_flag)
       {
-	  syslog(LOG_NOTICE, "from: %s ( %s ) for: %d, %d",
+	  syslog(LOG_INFO, "from: %s ( %s ) for: %d, %d",
 		 inet_ntoa(*faddr), gethost(faddr), lport, fport);
       }
 #endif
-    
-    if (debug_flag && syslog_flag)
-      syslog(LOG_DEBUG, "  After fscanf(), before k_getuid()");
-    
-    /*
-    ** Next - get the specific TCP connection and return the
-    ** uid - user number.
-    **
-    ** Try to fetch the information 5 times incase the
-    ** kernel changed beneath us and we missed or took
-    ** a fault.
-    */
-    for (try = 0;
-	 (try < 5 &&
-	   k_getuid(&faddr2, htons(fport), laddr, htons(lport), &uid) == -1);
-	 try++)
-      ;
 
-    if (try >= 5)
+    if (debug_flag && syslog_flag)
+      syslog(LOG_DEBUG, "  After fscanf(), before k_open()");
+
+
+    if (! k_opened)
+    {
+      /*
+      ** Open the kernel memory device and read the nlist table
+      ** 
+      ** Of course k_open should not call ERROR (which then exits)
+      ** but maybe use syslog(LOG_ERR) and return non-zero. But I am
+      ** too lazy to change them all ...
+      */
+      if (k_open() != 0)
+      {
+	if (syslog_flag) syslog(LOG_ERR, "k_open call failed");
+	printf("%d , %d : ERROR : %s\r\n",
+	     lport, fport,
+	     unknown_flag ? "UNKNOWN-ERROR" : "X-CANNOT-OPEN-KMEM");
+	continue;
+      }
+      k_opened = 1;
+    }
+
+
+    if (debug_flag && syslog_flag)
+      syslog(LOG_DEBUG, "  After k_open(), before k_getuid()");
+
+
+    /*
+    ** Get the specific TCP connection and return the uid - user number.
+    */
+
+#ifdef ALLOW_FORMAT
+    /* Initialize values, for architectures that do not set it */
+    pid = 0;
+    cmd = "";
+    cmd_and_args = "";
+#endif
+
+#define MAX_RETRY 20
+    /*
+    ** Try to fetch the information MAX_RETRY times in case the
+    ** kernel changed beneath us and we missed or took a fault.
+    **
+    ** Why would we ever fail? Is not there a reliable way for the
+    ** kernel to identify its sockets? Cannot we use that interface?
+    **
+    ** Used to be 5 times, but often this is not enough on Alpha OSF.
+    */
+/* #define SLEEP_BETWEEN_RETRIES 1 */
+    /*
+    ** If we failed in k_getuid, that is presumably because the OS was
+    ** busy creating or destroying processes. We may want to sleep for
+    ** a random time between retries, hoping for peace and quiet.
+    */
+
+/* k_getuid returns 0 on success, any non-zero on failure. */
+
+    for (try = 0;
+	 (try < MAX_RETRY &&
+	   k_getuid(&faddr2, htons(fport), laddr, htons(lport), &uid
+#ifdef ALLOW_FORMAT
+		    , &pid, &cmd, &cmd_and_args
+#endif
+		    ) != 0);
+	 try++)
+#ifdef SLEEP_BETWEEN_RETRIES
+      {
+	/* Seed the generator: lport should be unique (among other concurrent identd's) */
+	if (try < 1) srandom(lport);
+	/* This gives a max sleep of 0xffff = 65535 microsecs, about 32millisec average */
+	usleep(random()&0x00ffff);
+      }
+#else
+      ;
+#endif
+
+    if (try >= MAX_RETRY)
     {
       if (syslog_flag)
 	syslog(LOG_INFO, "Returned: %d , %d : NO-USER", lport, fport);
-      
+
       printf("%d , %d : ERROR : %s\r\n",
 	     lport, fport,
 	     unknown_flag ? "UNKNOWN-ERROR" : "NO-USER");
@@ -452,7 +526,7 @@ int parse(fp, laddr, faddr)
 
     if (try > 0 && syslog_flag)
       syslog(LOG_NOTICE, "k_getuid retries: %d", try);
-    
+
     if (debug_flag && syslog_flag)
       syslog(LOG_DEBUG, "  After k_getuid(), before getpwuid()");
 
@@ -461,8 +535,8 @@ int parse(fp, laddr, faddr)
     ** return it as an OTHER identifier
     */
     pwp = getpwuid(uid);
-    
-    if (!pwp)
+
+    if (!pwp || uid != pwp->pw_uid)
     {
       if (syslog_flag)
 	syslog(LOG_WARNING, "getpwuid() could not map uid (%d) to name",
@@ -476,12 +550,36 @@ int parse(fp, laddr, faddr)
       continue;
     }
 
+#ifdef ALLOW_FORMAT
+    grp = getgrgid(pwp->pw_gid);
+    if (!grp || pwp->pw_gid != grp->gr_gid)
+    {
+      if (syslog_flag)
+	syslog(LOG_WARNING, "getgrgid() could not map gid (%d) to name (for uid %d, name %s)",
+	       pwp->pw_gid, uid, pwp->pw_name);
+
+      printf("%d , %d : USERID : OTHER%s%s : %d\r\n",
+	     lport, fport,
+	     charset_name ? " , " : "",
+	     charset_name ? charset_name : "",
+	     uid);
+      continue;
+    }
+    sprintf (grname, "%.99s", grp->gr_name);
+#endif
+
     /*
     ** Hey! We finally made it!!!
     */
+#ifdef ALLOW_FORMAT
     if (syslog_flag)
-      syslog(LOG_INFO, "Successful lookup: %d , %d : %s\n",
+      syslog(LOG_DEBUG, "Successful lookup: %d , %d : %s.%s\n",
+	     lport, fport, pwp->pw_name, grname);
+#else
+    if (syslog_flag)
+      syslog(LOG_DEBUG, "Successful lookup: %d , %d : %s\n",
 	     lport, fport, pwp->pw_name);
+#endif
 
     if (noident_flag && check_noident(pwp->pw_dir))
     {
@@ -490,34 +588,98 @@ int parse(fp, laddr, faddr)
 	       pwp->pw_name,
 	       gethost(faddr),
 	       lport, fport);
-      
+
       printf("%d , %d : ERROR : HIDDEN-USER\r\n",
 	   lport, fport);
       continue;
     }
 
-    if (number_flag)
-      printf("%d , %d : USERID : OTHER%s%s : %d\r\n",
-	     lport, fport,
-	     charset_name ? " , " : "",
-	     charset_name ? charset_name : "",
-	     uid);
 #ifdef INCLUDE_CRYPT
-    else if (crypto_flag)
+    if (crypto_flag)
       printf("%d , %d : USERID : OTHER%s%s : [%s]\r\n",
 	     lport, fport,
 	     charset_name ? " , " : "",
 	     charset_name ? charset_name : "",
 	     make_packet (pwp->pw_uid, laddr, lport, faddr, fport));
-#endif
     else
-      printf("%d , %d : USERID : %s%s%s : %s\r\n",
+#endif
+#ifdef ALLOW_FORMAT
+    if (format_flag)
+    {
+      char* cp;
+      char** gmp;
+      long bp;
+      char buff[512];
+      for (cp = format, bp = 0; *cp != 0; cp++)
+      {
+	if (*cp == '%')
+	{
+	  cp++;
+	  if (*cp == 0) break;
+	  else if (*cp == 'u') sprintf (&buff[bp], "%.*s", 490-bp, pwp->pw_name);
+	  else if (*cp == 'U') sprintf (&buff[bp], "%d",           pwp->pw_uid);
+	  else if (*cp == 'g') sprintf (&buff[bp], "%.*s", 490-bp, grname);
+	  else if (*cp == 'G') sprintf (&buff[bp], "%d",           pwp->pw_gid);
+	  else if (*cp == 'c') sprintf (&buff[bp], "%.*s", 490-bp, cmd);
+	  else if (*cp == 'C') sprintf (&buff[bp], "%.*s", 490-bp, cmd_and_args);
+	  else if (*cp == 'l') {
+	    sprintf (&buff[bp], "%.*s", 490-bp, grname);
+	    bp += strlen(&buff[bp]); if (bp >= 490) break;
+	    setgrent();
+	    while (grp = getgrent()) {
+	      if (grp->gr_gid == pwp->pw_gid) continue;
+	      for (gmp = grp->gr_mem; *gmp && **gmp; gmp++) {
+		if (! strcmp(*gmp, pwp->pw_name)) {
+		  sprintf (&buff[bp], ",%.*s", 490-bp, grp->gr_name);
+		  bp += strlen(&buff[bp]);
+		  break;
+		}
+	      }
+	      if (bp >= 490) break;
+	    }
+	    endgrent();
+	  }
+	  else if (*cp == 'L') {
+	    sprintf (&buff[bp], "%d", pwp->pw_gid);
+	    bp += strlen(&buff[bp]); if (bp >= 490) break;
+	    setgrent();
+	    while (grp = getgrent()) {
+	      if (grp->gr_gid == pwp->pw_gid) continue;
+	      for (gmp = grp->gr_mem; *gmp && **gmp; gmp++) {
+		if (! strcmp(*gmp, pwp->pw_name)) {
+		  sprintf (&buff[bp], ",%d", grp->gr_gid);
+		  bp += strlen(&buff[bp]);
+		  break;
+		}
+	      }
+	      if (bp >= 490) break;
+	    }
+	    endgrent();
+	  }
+	  else if (*cp == 'p') sprintf (&buff[bp], "%d", pid);
+	  else { buff[bp] = *cp; buff[bp+1] = 0; }
+	  bp += strlen(&buff[bp]); if (bp >= 490) break;
+	}
+	else { buff[bp++] = *cp; if (bp >= 490) break; }
+      }
+      if (bp >= 490) { sprintf(&buff[490], "..."); bp = 493; }
+      buff[bp] = 0;
+      printf("%d , %d : USERID : %s%s%s :%s\r\n",
+	     lport, fport,
+	     other_flag ? "OTHER" : "UNIX",
+	     charset_name ? " , " : "",
+	     charset_name ? charset_name : "",
+	     buff);
+    }
+    else
+#endif
+      printf("%d , %d : USERID : %s%s%s :%s\r\n",
 	     lport, fport,
 	     other_flag ? "OTHER" : "UNIX",
 	     charset_name ? " , " : "",
 	     charset_name ? charset_name : "",
 	     pwp->pw_name);
-    
+
   } while(fflush(stdout), fflush(stderr), multi_flag && eat_whitespace());
 
   return 0;
